/**
 * Copyright (c) 2003, Brian McCallister
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name of the copyright holder nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 *THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package org.skife.intercept;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.*;

import net.sf.cglib.proxy.*;
import net.sf.cglib.core.MethodWrapper;
import net.sf.cglib.reflect.FastClass;

/**
 * Holds the information required for bulding intercepted instances.
 * This is really a factory for InterceptedObject instances.
 */
class InterceptedClass
{
    /** Builds new instances */
    private final FastClass fc;

    /** ?? */
    private final Map indexes = new HashMap();
    private final Method[] methods;

    /** means of building interceptors stored via [methodindex][wrapper array] */
    private final Wrapper[][] wrappers;

    /** the ordered array of callbacks */
    private final Callback[] callbacks;

    /** ?? */
    private final int targetIndex;

    /** arguments sent to constructor */
    private final Object[] nullArgs;

    /** which constructor do we use for building new wrapped objects */
    private int constructorIndex;

    /**
     * Create a new instance of an intercepted object wrapping base
     */
    public Object newInstance(final Object base)
    {
        try
        {
            Factory proxy = (Factory) fc.newInstance(constructorIndex, nullArgs);
            // use pre-defined method interceptors for this proxy
            proxy.setCallbacks(callbacks);
            // ...except for the InterceptedObject.getTarget method,
            // which we hardcode to return the base
            proxy.setCallback(targetIndex, new FixedValue()
            {
                public Object loadObject()
                {
                    return base;
                }
            });
            return proxy;
        }
        catch (InvocationTargetException e)
        {
            throw new RuntimeException("Should be impossible because we already tested for this when making the factory");
        }
    }

    /**
     * Traverse methods and cache which ones care about the interceptor
     */
    public void addInterceptor(Wrapper wrapper, Signature signature)
    {
        for (int i = 0; i < methods.length; i++)
        {
            if (signature.matches(methods[i]))
            {
                Object methodKey = MethodWrapper.create(methods[i]);
                int index = ((Integer) indexes.get(methodKey)).intValue();
                Wrapper[] old = wrappers[index];
                Wrapper[] temp = new Wrapper[old.length + 1];
                System.arraycopy(old, 0, temp, 0, old.length);
                temp[old.length] = wrapper;
                wrappers[index] = temp;
            }
        }
    }

    /**
     * Does the real work.
     */
    public InterceptedClass(Class targetClass)
    {
        methods = targetClass.getMethods();
        int count = methods.length;
        wrappers = new Wrapper[count][0];
        callbacks = new Callback[count + 2];
        Class[] callbackTypes = new Class[count + 2];

        final int noOpIndex = count;
        targetIndex = count + 1;

        Arrays.fill(callbackTypes, MethodInterceptor.class);
        callbackTypes[noOpIndex] = NoOp.class;
        callbackTypes[targetIndex] = FixedValue.class;

        for (int i = 0; i < count; i++)
        {
            final int index = i;
            indexes.put(MethodWrapper.create(methods[i]), new Integer(i));
            callbacks[i] = new MethodInterceptor()
            {
                public Object intercept(Object obj, Method method, Object[] args, MethodProxy proxy) throws Throwable
                {
                    // System.err.println("EXECUTING METHOD " + method + ", SLOT=" + slot);
                    Object target = ((InterceptedObject) obj).getTarget(null);
                    Wrapper[] list = wrappers[index];
                    int size = list.length;
                    if (size == 0)
                    {
                        // optimization for public methods that don't have any interceptors attached
                        return proxy.invoke(target, args);
                    }
                    else
                    {
                        Interceptor[] intercepts = new Interceptor[size];
                        for (int j = 0; j < size; j++)
                        {
                            intercepts[j] = list[j].getIntercept();
                        }
                        return new InterceptorStack(intercepts, target, method, args, proxy).yield();
                    }
                }
            };
        }
        // no-op callback is used for all package and protected methods
        callbacks[noOpIndex] = NoOp.INSTANCE;

        // this interceptor will be replaced whenever a new instance is created.
        // we use a dummy FixedValue because callback types cannot change
        callbacks[targetIndex] = new FixedValue()
        {
            public Object loadObject()
            {
                return null;
            }
        };

        Enhancer e = new Enhancer();

        Class targetSuper = targetClass.getSuperclass();
        if (targetSuper != null && targetSuper.getName().equals("java.lang.reflect.Proxy"))
        {
            Class[] c1 = targetClass.getInterfaces();
            Class[] c2 = new Class[c1.length + 1];
            System.arraycopy(c1, 0, c2, 0, c1.length);
            c2[c2.length - 1] = InterceptedObject.class;
            e.setInterfaces(c2);
        }
        else
        {
            e.setSuperclass(targetClass);
            e.setInterfaces(new Class[]{InterceptedObject.class});
        }
        e.setCallbackTypes(callbackTypes);
        e.setCallbackFilter(new CallbackFilter()
        {
            public int accept(Method method)
            {
                Integer index = (Integer) indexes.get(MethodWrapper.create(method));
                if (index != null)
                {
                    return index.intValue();
                }
                else if (method.getDeclaringClass().equals(InterceptedObject.class))
                {
                    return targetIndex;
                }
                else
                {
                    return noOpIndex;
                }
            }

            // overriding equals and hashCode is important so that when this
            // same InterceptedClass is created in another broker the same underlying
            // generated class will be found in the cache by CGLIB
            public boolean equals(Object o)
            {
                return o != null && o.getClass().equals(getClass());
            }

            public int hashCode()
            {
                return 2334; // any constant value will do
            }
        });

        fc = FastClass.create(e.createClass());
        try
        {
            int numArgs = 0;
            // look for a no-arg constructor
            constructorIndex = fc.getIndex(new Class[0]);
            if (constructorIndex < 0)
            {
                // Doesn't have a no-arg constructor, maybe it is an inner class?
                // Choose constructor with least arguments
                int best = Integer.MAX_VALUE;
                Constructor[] constructors = targetClass.getDeclaredConstructors();
                for (int i = 0; i < constructors.length; i++)
                {
                    Class[] types = constructors[i].getParameterTypes();
                    int index = fc.getIndex(types);
                    if (types.length < best && index >= 0)
                    {
                        constructorIndex = index;
                        numArgs = best = types.length;
                        if (types.length == 1)
                        {
                            // we're not going to get smaller, so use this one
                            break;
                        }
                    }
                }
            }
            nullArgs = new Object[numArgs];
            // test newInstance so we can catch errors earlier
            // passing all nulls to constructor may cause errors for some classes
            fc.newInstance(constructorIndex, nullArgs);
        }
        catch (InvocationTargetException ex)
        {
            throw new RuntimeException("TODO! proxied class has thrown an exception from it's constructor: " + ex);
        }
    }
}
