/**
 * Copyright (c) 2003, Brian McCallister
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name of the copyright holder nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 *THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package org.skife.intercept;

import net.sf.cglib.reflect.FastClass;

import java.lang.reflect.InvocationTargetException;

/**
 * provide interception matches against arbitrary non-final objects
 */
public class InterceptionBroker
{
    /**
     * Actual operations will be delegated to this.
     */
    private Broker broker = null;

    /**
     * Create a new, unbaked InterceptionBroker
     */
    public InterceptionBroker()
    {
        this.broker = new DynamicBroker();
    }
    
    /**
     * Wrap an arbitrary object in whatever interceptors match it.
     * <p>
     * It is important to note that an object can only be wrapped
     * by a single broker at a time, or to put it more specifically,
     * a broker will NOT wrap an already wrapped object -- this is
     * a meaningless triviality when attempting to wrap the same object
     * via the same broker multiple times, but it means that if an
     * InterceptedObject is passed to this method, you will get the same
     * object back with no changes.
     *
     * @param obj must not be declared final
     * @return a subclass wrapping obj which will be intercepted
     */
    public InterceptedObject wrap(Object obj)
    {
        return broker.wrap(obj);
    }

    /**
     * Bake this broker. Baking removes the ability to add new interceptors
     * but significantly increases the performance of objects wrapped
     * after baking has occured.
     */
    public synchronized void bake()
    {
        this.broker = new BakedBroker(this.broker);
    }

    /**
     * Used to add an interceptor where a new <code>Interceptor</code>
     * will be instantiated for each invocation.
     *
     * @param interceptor Class implementing <code>Interceptor</code>
     * @param signature Signature which will be matched for interception
     * @throws UnsupportedOperationException if the broker has been baked
     */
    public synchronized void addInterceptor(Class interceptor, Signature signature)
    {
        broker.addInterceptor(new NewInstanceWrapper(interceptor), signature);
    }

     /**
     * Used to add an interceptor where a new <code>Interceptor</code>
     * will be instantiated for each invocation.
     *
     * @param factory used to build a new Interceptor on each invocation
     * @param signature Signature which will be matched for interception
     * @throws UnsupportedOperationException if the broker has been baked
     */
    public synchronized void addInterceptor(InterceptorFactory factory, Signature signature)
    {
        broker.addInterceptor(new FactoryWrapper(factory), signature);
    }

    /**
     * The <code>Interceptor</code> passed here will be used for every method
     * invocation. Make sure it is threadsafe if you are in a multithreaded environment!
     *
     * @param interceptor Singleton interceptor
     * @param signature
     * @throws UnsupportedOperationException if the broker has been baked
     */
    public synchronized void addInterceptor(Interceptor interceptor, Signature signature)
    {
        broker.addInterceptor(new SingletonWrapper(interceptor), signature);
    }

    private static class SingletonWrapper
    implements Wrapper
    {
        private final Interceptor singleton;

        SingletonWrapper(Interceptor intercept)
        {
            this.singleton = intercept;
        }

        public Interceptor getIntercept()
        {
            return this.singleton;
        }
    }

    private static class NewInstanceWrapper
    implements Wrapper
    {
        private final FastClass fc;

        NewInstanceWrapper(Class clazz)
        {
            fc = FastClass.create(clazz);
        }

        public Interceptor getIntercept()
        throws InvocationTargetException
        {
            return (Interceptor)fc.newInstance();
        }
    }

    private static class FactoryWrapper implements Wrapper
    {
        private InterceptorFactory factory;

        FactoryWrapper(InterceptorFactory factory)
        {
            this.factory = factory;
        }

        public Interceptor getIntercept() throws Exception
        {
            return factory.build();
        }

    }
}
