/**
 * Copyright (c) 2003, Brian McCallister
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name of the copyright holder nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 *THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package org.skife.intercept.interceptors;

import org.skife.intercept.Interceptor;
import org.skife.intercept.InterceptorStack;

import java.lang.reflect.Method;

/**
 * Throws an <code>UnsupportedOperationException</code>. Typically used
 * with an interceptor to prevent specific calls, for instance:
 * <pre><code>
 * InterceptionBroker broker = new InterceptionBroker();
 * broker.addInterceptor(new BlockingInterceptor("Cannot modify locked JavaBean"), new Signature()
 * {
 *     public boolean matches(Method m)
 *     {
 *         return (m.getName().startsWith("set") ? true : false);
 *     }
 * });
 *
 * // Create a wrapped, javabean-setter blocked version of foo. Note that foo can
 * // still be modified, but unmodifiableFoo cannot have setXxx(...) called
 * Foo unmodifiableFoo = (Foo) broker.wrap(foo);
 * </code></pre>
 */
public class BlockingInterceptor implements Interceptor
{
    /**
     * exception message
     */
    private String msg;

    /**
     * Create a new BlockingInterceptor which will throw an exception with the
     * message set to the argument
     *
     * @param msg Message for the <code>UnsupportedOperationException</code>
     */
    public BlockingInterceptor(String msg)
    {
        this.msg = msg;
    }

    /**
     * Throws an <code>UnsupportedOperationException</code>
     */
    public Object intercept(Object base, Method method, Object[] args, InterceptorStack stack)
            throws Throwable
    {
        throw new UnsupportedOperationException(msg);
    }
}
