/**
 * Copyright (c) 2003, Brian McCallister
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name of the copyright holder nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 *THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package org.skife.intercept.interceptors.tx;

import org.skife.intercept.Interceptor;
import org.skife.intercept.InterceptorStack;

import java.lang.reflect.Method;

/**
 * Provides container managed transactions for arbitrary containers
 * and transaction types.
 * <p>
 * If the intercepted class responds to <code>setTransaction(Transaction)</code>
 * than the current transaction will be populated onto it before the stack is
 * yielded.
 *
 * @see TransactionFactory
 * @see Transaction
 */
public class TransactionInterceptor implements Interceptor
{
    private TransactionFactory factory;

    /**
     * Designed for Type 2 IoC - setTransactionFactory(...) <b>must</b> be called
     */
    public TransactionInterceptor()
    {
        // Nothing
    }

    /**
     * Designed for Type 3 IoC
     *
     * @param factory Transactionfactory to use.
     */
    public TransactionInterceptor(TransactionFactory factory)
    {
        this.factory = factory;
    }

    /**
     * Specify the transaction factory implementation to use
     *
     * @param factory
     */
    public void setTransactionFactory(TransactionFactory factory)
    {
        this.factory = factory;
    }

    /**
     * Do the work
     */
    public Object intercept(Object base, Method method, Object[] args, InterceptorStack stack)
            throws Throwable
    {
        Transaction tx = factory.build();
        if (!tx.isOpen()) tx.begin();
        try
        {
            // TODO I hate Sun's decision to throw excpetion instead of return null
            Method st = base.getClass().getMethod("setTransaction", new Class[]{Transaction.class});
            st.invoke(base, new Object[]{tx});
        }
        catch (Exception e)
        {
        }
        Object value = null;
        try
        {
            value = stack.yield();
        }
        catch (Throwable t)
        {
            if (tx.isOpen()) tx.rollback();
            throw t;
        }
        if (tx.isOpen()) tx.commit();
        return value;
    }
}
