/**
 * Copyright (c) 2003, Brian McCallister
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name of the copyright holder nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package org.skife.intercept;

import com.mockobjects.dynamic.Mock;
import com.mockobjects.dynamic.AnyConstraintMatcher;
import junit.framework.Assert;
import junit.framework.TestCase;
import org.skife.intercept.interceptors.InefficientLoggingInterceptor;
import org.skife.intercept.signatures.StrictSignature;
import org.skife.intercept.signatures.NamedSignature;
import org.skife.intercept.signatures.EverythingSignature;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.HashMap;
import java.util.Map;

public class TestInterceptionBroker extends TestCase
{
    InterceptionBroker broker;

    public void setUp()
    {
        broker = new InterceptionBroker();
    }

    public void testWrapBeforeDeclaringInterceptors() throws Exception
    {
        MockInterceptor mi = new MockInterceptor();
        HashMap map = new HashMap();
        Map same = (Map) broker.wrap(map);
        broker.addInterceptor(mi, new StrictSignature(map.getClass(),
                                                      "put",
                                                      new Class[]{Object.class, Object.class},
                                                      Object.class));

        same.put("thing", "value");
        Assert.assertTrue(mi.value);
        Assert.assertNotNull(same.get("thing"));
    }

    public void testInterceptorFactory() throws Exception
    {
        final MockInterceptor one = new MockInterceptor();
        final MockInterceptor two = new MockInterceptor();
        InterceptorFactory factory = new InterceptorFactory()
        {
            private boolean first = true;

            public Interceptor build()
            {
                Interceptor foo = (first ? one : two);
                if (first) first = false;
                return foo;
            }
        };
        broker.addInterceptor(factory, new NamedSignature("put"));
        HashMap foo = (HashMap) broker.wrap(new HashMap());
        foo.put("a", "1");
        assertTrue(one.value);
        assertFalse(two.value);
        foo.put("b", "2");
        assertTrue(two.value);
    }

    public void testWrapAfterDeclaringIntercepts() throws Exception
    {
        MockInterceptor mi = new MockInterceptor();
        broker.addInterceptor(mi, new StrictSignature(HashMap.class,
                                                      "put",
                                                      new Class[]{Object.class, Object.class},
                                                      Object.class));
        HashMap map = new HashMap();
        Map same = (Map) broker.wrap(map);
        same.put("thing", "value");
        Assert.assertTrue(mi.value);
        Assert.assertNotNull(same.get("thing"));
    }

    public void testWrapInterceptedObject() throws Exception
    {
        HashMap one = new HashMap();
        HashMap wrapped = (HashMap) broker.wrap(one);
        MockInterceptor mock = new MockInterceptor();
        broker.addInterceptor(mock, new Signature()
        {
            public boolean matches(Method m)
            {
                return true;
            }
        });
        HashMap again = (HashMap) broker.wrap(wrapped);
        again.put("a", "b");
        assertTrue(mock.value);
    }

    public void testWrapAnonymousInnerClass() throws Exception
    {
        MockInterceptor mock = new MockInterceptor();
        broker.addInterceptor(mock, new Signature()
        {
            public boolean matches(Method m)
            {
                return m.getName().equals("toString");
            }
        });

        Object foo = new Object()
        {
            public String toString()
            {
                return "foo";
            };
        };

        Object wrapped = null;
        try
        {
            wrapped = broker.wrap(foo);
        }
        catch (Exception e)
        {
            System.err.println(e.getMessage());
            e.printStackTrace(System.err);
            fail("Failed trying to wrap anonymous inner class");
        }
        wrapped.toString();
        assertTrue(mock.value);
    }

    public void testWrapInnerClass() throws Exception
    {
        MockInterceptor mock = new MockInterceptor();
        broker.addInterceptor(mock, new Signature()
        {
            public boolean matches(Method m)
            {
                return m.getName().equals("toString");
            }
        });

        InnerClass foo = new InnerClass();

        Object wrapped = null;
        try
        {
            wrapped = broker.wrap(foo);
        }
        catch (Exception e)
        {
            System.err.println(e.getMessage());
            e.printStackTrace(System.err);
            fail("Failed trying to wrap inner class");
        }
        wrapped.toString();
        assertTrue(mock.value);
    }

    class InnerClass
    {
        public String toString()
        {
            return "foo";
        }
    }


    /**
     * This is really a test of wrapping a Proxy
     */
    public void testWrapDynaMock()
    {
        Mock mock = new Mock(Map.class);
        mock.expect("put", new AnyConstraintMatcher());
        InterceptionBroker broker = new InterceptionBroker();
        Map wrap = (Map) broker.wrap(mock.proxy());
        wrap.put("1", "2");
        mock.verify();
    }

    public void testWrapDynamicProxy() throws Exception
    {
        final HashMap base = new HashMap();
        InvocationHandler handler = new InvocationHandler()
        {
            public Object invoke(Object proxy, Method method, Object[] args)
                    throws Throwable
            {
                return method.invoke(base, args);
            }

        };
        Map map = (Map) Proxy.newProxyInstance(this.getClass().getClassLoader(),
                                               new Class[]{Map.class},
                                               handler);
        map.put("1", "1");

        InterceptionBroker broker = new InterceptionBroker();
        try
        {
            Map wrap = (Map) broker.wrap(map);
        }
        catch (NullPointerException e)
        {
            System.err.println(e.getMessage());
            e.printStackTrace(System.err);
            fail("Null pointer exception is thrown when wrapping a java.lang.reflect.proxy");
        }
        assertTrue("Should get here =)", true);
    }

    public void _testOrder1()
    {
        HashMap base = new HashMap();
        //Map wrap = (Map) broker.wrap(base);

        // called last
        broker.addInterceptor(new Interceptor()
        {
            public Object intercept(Object base, Method method, Object[] args, InterceptorStack stack) throws Throwable
            {
                System.err.println("1");
                return stack.yield();
            }
        }, new EverythingSignature());

        //called second
        broker.addInterceptor(new Interceptor()
        {
            public Object intercept(Object base, Method method, Object[] args, InterceptorStack stack) throws Throwable
            {
                System.err.println("2");
                return stack.yield();
            }
        }, new EverythingSignature());

        //called first
        broker.addInterceptor(new Interceptor()
        {
            public Object intercept(Object base, Method method, Object[] args, InterceptorStack stack) throws Throwable
            {
                System.err.println("3");
                return stack.yield();
            }
        }, new EverythingSignature());

        broker.addInterceptor(new Interceptor()
        {
            public Object intercept(Object base, Method method, Object[] args, InterceptorStack stack) throws Throwable
            {
                System.err.println(4);
                return stack.yield();
            }
        }, new EverythingSignature());
        broker.addInterceptor(new Interceptor()
        {
            public Object intercept(Object base, Method method, Object[] args, InterceptorStack stack) throws Throwable
            {
                System.err.println(5);
                return stack.yield();
            }
        }, new EverythingSignature());
        broker.addInterceptor(new Interceptor()
        {
            public Object intercept(Object base, Method method, Object[] args, InterceptorStack stack) throws Throwable
            {
                System.err.println(6);
                return stack.yield();
            }
        }, new EverythingSignature());
        broker.addInterceptor(new Interceptor()
        {
            public Object intercept(Object base, Method method, Object[] args, InterceptorStack stack) throws Throwable
            {
                System.err.println(7);
                return stack.yield();
            }
        }, new EverythingSignature());
        broker.addInterceptor(new Interceptor()
        {
            public Object intercept(Object base, Method method, Object[] args, InterceptorStack stack) throws Throwable
            {
                System.err.println(8);
                return stack.yield();
            }
        }, new EverythingSignature());
        Map wrap = (Map) broker.wrap(base);
        wrap.put("1", "2");
    }
}
